@@ -27,8 +27,7 @@ use pact_models::generators::GeneratorTestMode;
 use pact_models::http_parts::HttpPart;
 use pact_models::pact::Pact;
 use pact_models::query_strings::parse_query_string;
-use pact_models::request::Request;
-use pact_models::sync_pact::RequestResponsePact;
+use pact_models::v4::http_parts::HttpRequest;
 
 use crate::matching::{match_request, MatchResult};
 use crate::mock_server::MockServer;
@@ -89,15 +88,15 @@ fn extract_headers(headers: &hyper::HeaderMap) -> Result<Option<HashMap<String,
   }
 }
 
-fn extract_body(bytes: bytes::Bytes, request: &Request) -> OptionalBody {
+fn extract_body(bytes: bytes::Bytes, request: &HttpRequest) -> OptionalBody {
     if bytes.len() > 0 {
-      OptionalBody::Present(bytes, request.content_type())
+      OptionalBody::Present(bytes, request.content_type(), None)
     } else {
       OptionalBody::Empty
     }
 }
 
-async fn hyper_request_to_pact_request(req: hyper::Request<Body>) -> Result<Request, InteractionError> {
+async fn hyper_request_to_pact_request(req: hyper::Request<Body>) -> Result<HttpRequest, InteractionError> {
     let method = req.method().to_string();
     let path = extract_path(req.uri());
     let query = extract_query_string(req.uri());
@@ -107,15 +106,15 @@ async fn hyper_request_to_pact_request(req: hyper::Request<Body>) -> Result<Requ
         .await
         .map_err(|_| InteractionError::RequestBodyError)?;
 
-    let request = Request {
+    let request = HttpRequest {
       method,
       path,
       query,
       headers,
-      .. Request::default()
+      .. HttpRequest::default()
     };
 
-    Ok(Request {
+    Ok(HttpRequest {
       body: extract_body(body_bytes, &request),
       .. request.clone()
     })
@@ -149,32 +148,33 @@ fn set_hyper_headers(builder: &mut ResponseBuilder, headers: &Option<HashMap<Str
     Ok(())
 }
 
-fn error_body(request: &Request, error: &String) -> String {
+fn error_body(request: &HttpRequest, error: &String) -> String {
     let body = json!({ "error" : format!("{} : {:?}", error, request) });
     body.to_string()
 }
 
-fn match_result_to_hyper_response(
-  request: &Request,
+async fn match_result_to_hyper_response(
+  request: &HttpRequest,
   match_result: MatchResult,
   mock_server: Arc<Mutex<MockServer>>
 ) -> Result<Response<Body>, InteractionError> {
-  let cors_preflight = {
+  let (context, cors_preflight) = {
     let ms = mock_server.lock().unwrap();
-    ms.config.cors_preflight
-  };
-
-  match match_result {
-    MatchResult::RequestMatch(_, ref response) => {
-      let ms = mock_server.lock().unwrap();
-      let context = hashmap!{
+    (
+      hashmap!{
         "mockServer" => json!({
           "href": ms.url(),
           "port": ms.port
         })
-      };
+      },
+      ms.config.cors_preflight
+    )
+  };
+
+  match match_result {
+    MatchResult::RequestMatch(_, ref response) => {
       debug!("Test context = {:?}", context);
-      let response = pact_matching::generate_response(response, &GeneratorTestMode::Consumer, &context);
+      let response = pact_matching::generate_response(response, &GeneratorTestMode::Consumer, &context).await;
       info!("Request matched, sending response {}", response);
       if response.has_text_body() {
         debug!("     body: '{}'", response.body.str_value());
@@ -190,7 +190,7 @@ fn match_result_to_hyper_response(
       set_hyper_headers(&mut builder, &response.headers)?;
 
       builder.body(match response.body {
-        OptionalBody::Present(ref s, _) => Body::from(s.clone()),
+        OptionalBody::Present(ref s, _, _) => Body::from(s.clone()),
         _ => Body::empty()
       })
         .map_err(|_| InteractionError::ResponseBodyError)
@@ -235,7 +235,7 @@ fn match_result_to_hyper_response(
 
 async fn handle_request(
   req: hyper::Request<Body>,
-  pact: Arc<RequestResponsePact>,
+  pact: Arc<Mutex<dyn Pact + Send + Sync>>,
   matches: Arc<Mutex<Vec<MatchResult>>>,
   mock_server: Arc<Mutex<MockServer>>
 ) -> Result<Response<Body>, InteractionError> {
@@ -253,11 +253,15 @@ async fn handle_request(
     debug!("     body: '{}'", pact_request.body.str_value());
   }
 
-  let match_result = match_request(&pact_request, pact.interactions());
+  let pact = {
+    let inner = pact.lock().unwrap();
+    inner.as_v4_pact().unwrap()
+  };
+  let match_result = match_request(&pact_request, &pact).await;
 
   matches.lock().unwrap().push(match_result.clone());
 
-  match_result_to_hyper_response(&pact_request, match_result, mock_server)
+  match_result_to_hyper_response(&pact_request, match_result, mock_server).await
 }
 
 // TODO: Should instead use some form of X-Pact headers
@@ -289,14 +293,14 @@ fn handle_mock_request_error(result: Result<Response<Body>, InteractionError>) -
 // The reason that the function itself is still async (even if it performs
 // no async operations) is that it needs a tokio context to be able to call try_bind.
 pub(crate) async fn create_and_bind(
-  pact: RequestResponsePact,
+  pact: Arc<Mutex<dyn Pact + Send + Sync>>,
   addr: SocketAddr,
   shutdown: impl std::future::Future<Output = ()>,
   matches: Arc<Mutex<Vec<MatchResult>>>,
   mock_server: Arc<Mutex<MockServer>>,
   mock_server_id: &String
 ) -> Result<(impl std::future::Future<Output = ()>, SocketAddr), hyper::Error> {
-  let pact = Arc::new(pact);
+  let pact = pact.clone();
   let ms_id = Arc::new(mock_server_id.clone());
 
   let server = Server::try_bind(&addr)?
@@ -306,7 +310,7 @@ pub(crate) async fn create_and_bind(
       let mock_server = mock_server.clone();
       let mock_server_id = ms_id.clone();
 
-      LOG_ID.scope(mock_server_id.to_string(), async {
+      LOG_ID.scope(mock_server_id.to_string(), async move {
         Ok::<_, hyper::Error>(
           service_fn(move |req| {
             let pact = pact.clone();
@@ -314,7 +318,7 @@ pub(crate) async fn create_and_bind(
             let mock_server = mock_server.clone();
             let mock_server_id = mock_server_id.clone();
 
-            LOG_ID.scope(mock_server_id.to_string(), async {
+            LOG_ID.scope(mock_server_id.to_string(), async move {
               handle_mock_request_error(
                 handle_request(req, pact, matches, mock_server).await
               )
@@ -355,15 +359,13 @@ impl hyper::server::accept::Accept for HyperAcceptor {
 }
 
 pub(crate) async fn create_and_bind_tls(
-  pact: RequestResponsePact,
+  pact: Arc<Mutex<dyn Pact + Send + Sync>>,
   addr: SocketAddr,
   shutdown: impl std::future::Future<Output = ()>,
   matches: Arc<Mutex<Vec<MatchResult>>>,
   tls_cfg: ServerConfig,
   mock_server: Arc<Mutex<MockServer>>
 ) -> Result<(impl std::future::Future<Output = ()>, SocketAddr), io::Error> {
-  let pact = Arc::new(pact);
-
   let tcp = TcpListener::bind(&addr).await?;
   let socket_addr = tcp.local_addr()?;
   let tls_acceptor = Arc::new(TlsAcceptor::from(Arc::new(tls_cfg)));
@@ -421,6 +423,8 @@ mod tests {
   use hyper::header::{ACCEPT, CONTENT_TYPE, USER_AGENT};
   use hyper::HeaderMap;
 
+  use pact_models::prelude::RequestResponsePact;
+
   use super::*;
 
   #[tokio::test]
@@ -429,7 +433,7 @@ mod tests {
     let matches = Arc::new(Mutex::new(vec![]));
 
     let (future, _) = create_and_bind(
-      RequestResponsePact::default(),
+      RequestResponsePact::default().thread_safe(),
       ([0, 0, 0, 0], 0 as u16).into(),
       async {
           shutdown_rx.await.ok();
