@@ -10,13 +10,16 @@ use std::path::PathBuf;
 use std::sync::{Arc, Mutex};
 
 use log::*;
+use pact_plugin_driver::plugin_manager::drop_plugin_access;
+use pact_plugin_driver::plugin_models::{PluginDependency, PluginDependencyType};
 use rustls::ServerConfig;
 use serde::{Deserialize, Serialize};
 use serde_json::json;
 
 use pact_models::pact::{Pact, write_pact};
-use pact_models::request::Request;
+use pact_models::PactSpecification;
 use pact_models::sync_pact::RequestResponsePact;
+use pact_models::v4::http_parts::HttpRequest;
 
 use crate::hyper_server;
 use crate::matching::MatchResult;
@@ -25,7 +28,9 @@ use crate::matching::MatchResult;
 #[derive(Debug, Default, Clone)]
 pub struct MockServerConfig {
   /// If CORS Pre-Flight requests should be responded to
-  pub cors_preflight: bool
+  pub cors_preflight: bool,
+  /// Pact specification to use
+  pub pact_specification: PactSpecification
 }
 
 /// Mock server scheme
@@ -81,14 +86,16 @@ pub struct MockServer {
   /// Mock server config
   pub config: MockServerConfig,
   /// Metrics collected by the mock server
-  pub metrics: MockServerMetrics
+  pub metrics: MockServerMetrics,
+  /// Pact spec version to use
+  pub spec_version: PactSpecification
 }
 
 impl MockServer {
   /// Create a new mock server, consisting of its state (self) and its executable server future.
   pub async fn new(
     id: String,
-    pact: Box<dyn Pact>,
+    pact: Box<dyn Pact + Send + Sync>,
     addr: std::net::SocketAddr,
     config: MockServerConfig
   ) -> Result<(Arc<Mutex<MockServer>>, impl std::future::Future<Output = ()>), String> {
@@ -105,11 +112,12 @@ impl MockServer {
       matches: matches.clone(),
       shutdown_tx: RefCell::new(Some(shutdown_tx)),
       config: config.clone(),
-      metrics: MockServerMetrics::default()
+      metrics: MockServerMetrics::default(),
+      spec_version: pact_specification(config.pact_specification, pact.specification_version())
     }));
 
     let (future, socket_addr) = hyper_server::create_and_bind(
-      pact.as_request_response_pact().unwrap(),
+      pact.thread_safe(),
       addr,
       async {
         shutdown_rx.await.ok();
@@ -152,11 +160,12 @@ impl MockServer {
       matches: matches.clone(),
       shutdown_tx: RefCell::new(Some(shutdown_tx)),
       config: config.clone(),
-      metrics: MockServerMetrics::default()
+      metrics: MockServerMetrics::default(),
+      spec_version: pact_specification(config.pact_specification, pact.specification_version())
     }));
 
     let (future, socket_addr) = hyper_server::create_and_bind_tls(
-      pact.as_request_response_pact().unwrap(),
+      pact.thread_safe(),
       addr,
       async {
         shutdown_rx.await.ok();
@@ -179,6 +188,17 @@ impl MockServer {
 
   /// Send the shutdown signal to the server
   pub fn shutdown(&mut self) -> Result<(), String> {
+    // Need to check if any plugins need to be shutdown
+    let pact = self.pact.lock().unwrap();
+    for plugin in pact.plugin_data() {
+      let dependency = PluginDependency {
+        name: plugin.name,
+        version: Some(plugin.version),
+        dependency_type: PluginDependencyType::Plugin
+      };
+      drop_plugin_access(&dependency);
+    }
+
     let shutdown_future = &mut *self.shutdown_tx.borrow_mut();
     match shutdown_future.take() {
       Some(sender) => {
@@ -219,7 +239,7 @@ impl MockServer {
       let mismatches = matches.iter()
         .filter(|m| !m.matched() && !m.cors_preflight())
         .map(|m| m.clone());
-      let requests: Vec<Request> = matches.iter().map(|m| {
+      let requests: Vec<HttpRequest> = matches.iter().map(|m| {
         match m {
           MatchResult::RequestMatch(request, _) => Some(request),
           MatchResult::RequestMismatch(request, _) => Some(request),
@@ -231,7 +251,7 @@ impl MockServer {
       let pact = self.pact.lock().unwrap();
       let interactions = pact.interactions();
       let missing = interactions.iter()
-        .map(|i| i.as_request_response().unwrap().request)
+        .map(|i| i.as_v4_http().unwrap().request)
         .filter(|req| !requests.contains(req))
         .map(|req| MatchResult::MissingRequest(req.clone()));
       mismatches.chain(missing).collect()
@@ -239,7 +259,9 @@ impl MockServer {
 
   /// Mock server writes its pact out to the provided directory
   pub fn write_pact(&self, output_path: &Option<String>, overwrite: bool) -> anyhow::Result<()> {
-    let pact = self.pact.lock().unwrap().boxed();
+    trace!("write_pact: output_path = {:?}, overwrite = {}", output_path, overwrite);
+    let mut pact = self.pact.lock().unwrap();
+    pact.add_md_version("mockserver", option_env!("CARGO_PKG_VERSION").unwrap_or("unknown"));
     let pact_file_name = pact.default_file_name();
     let filename = match *output_path {
       Some(ref path) => {
@@ -251,8 +273,11 @@ impl MockServer {
     };
 
     info!("Writing pact out to '{}'", filename.display());
-    let specification = pact.specification_version();
-    match write_pact(pact, filename.as_path(), specification, overwrite) {
+    let specification = match self.spec_version {
+      PactSpecification::Unknown => PactSpecification::V3,
+      _ => self.spec_version
+    };
+    match write_pact(pact.boxed(), filename.as_path(), specification, overwrite) {
       Ok(_) => Ok(()),
       Err(err) => {
         warn!("Failed to write pact to file - {}", err);
@@ -272,6 +297,13 @@ impl MockServer {
     }
 }
 
+fn pact_specification(spec1: PactSpecification, spec2: PactSpecification) -> PactSpecification {
+  match spec1 {
+    PactSpecification::Unknown => spec2,
+    _ => spec1
+  }
+}
+
 impl Clone for MockServer {
   /// Make a clone all of the MockServer fields.
   /// Note that the clone of the original server cannot be shut down directly.
@@ -286,7 +318,8 @@ impl Clone for MockServer {
       matches: self.matches.clone(),
       shutdown_tx: RefCell::new(None),
       config: self.config.clone(),
-      metrics: self.metrics.clone()
+      metrics: self.metrics.clone(),
+      spec_version: self.spec_version
     }
   }
 }
@@ -303,7 +336,8 @@ impl Default for MockServer {
       matches: Arc::new(Mutex::new(vec![])),
       shutdown_tx: RefCell::new(None),
       config: Default::default(),
-      metrics: Default::default()
+      metrics: Default::default(),
+      spec_version: Default::default()
     }
   }
 }
